@@ -1,76 +1,67 @@
-from __future__ import absolute_import, print_function, division, unicode_literals
-
-import _io
 import inspect
 import json as json_module
 import logging
-import re
-from itertools import groupby
-
-import six
-
 from collections import namedtuple
-from functools import update_wrapper
-from requests.adapters import HTTPAdapter
-from requests.exceptions import ConnectionError
-from requests.sessions import REDIRECT_STATI
-from requests.utils import cookiejar_from_dict
-from responses.matchers import json_params_matcher as _json_params_matcher
-from responses.matchers import urlencoded_params_matcher as _urlencoded_params_matcher
+from collections.abc import Sequence
+from collections.abc import Sized
+from functools import wraps
+from http import client
+from http import cookies
+from itertools import groupby
+from re import Pattern
 from warnings import warn
 
-try:
-    from collections.abc import Sequence, Sized
-except ImportError:
-    from collections import Sequence, Sized
+from requests.adapters import HTTPAdapter
+from requests.exceptions import ConnectionError
+from requests.utils import cookiejar_from_dict
+
+from responses.matchers import json_params_matcher as _json_params_matcher
+from responses.matchers import query_string_matcher as _query_string_matcher
+from responses.matchers import urlencoded_params_matcher as _urlencoded_params_matcher
+from responses.registries import FirstMatchRegistry
 
 try:
     from requests.packages.urllib3.response import HTTPResponse
-except ImportError:
-    from urllib3.response import HTTPResponse
+except ImportError:  # pragma: no cover
+    from urllib3.response import HTTPResponse  # pragma: no cover
 try:
     from requests.packages.urllib3.connection import HTTPHeaderDict
-except ImportError:
-    from urllib3.response import HTTPHeaderDict
+except ImportError:  # pragma: no cover
+    from urllib3.response import HTTPHeaderDict  # pragma: no cover
 try:
     from requests.packages.urllib3.util.url import parse_url
-except ImportError:
-    from urllib3.util.url import parse_url
+except ImportError:  # pragma: no cover
+    from urllib3.util.url import parse_url  # pragma: no cover
 
-if six.PY2:
-    from urlparse import urlparse, parse_qsl, urlsplit, urlunsplit
-    from urllib import quote
-else:
-    from urllib.parse import urlparse, parse_qsl, urlsplit, urlunsplit, quote
-
-if six.PY2:
-    try:
-        from six import cStringIO as BufferIO
-    except ImportError:
-        from six import StringIO as BufferIO
-else:
-    from io import BytesIO as BufferIO
-
-try:
-    from unittest import mock as std_mock
-except ImportError:
-    import mock as std_mock
-
-try:
-    Pattern = re._pattern_type
-except AttributeError:
-    # Python 3.7
-    Pattern = re.Pattern
-
-UNSET = object()
+from io import BufferedReader
+from io import BytesIO
+from unittest import mock as std_mock
+from urllib.parse import parse_qsl
+from urllib.parse import quote
+from urllib.parse import urlsplit
+from urllib.parse import urlunparse
+from urllib.parse import urlunsplit
 
 Call = namedtuple("Call", ["request", "response"])
-
 _real_send = HTTPAdapter.send
+_UNSET = object()
 
 logger = logging.getLogger("responses")
 
 
+class FalseBool:
+    """Class to mock up built-in False boolean.
+
+    Used for backwards compatibility, see
+    https://github.com/getsentry/responses/issues/464
+    """
+
+    def __bool__(self):
+        return False
+
+    __nonzero__ = __bool__
+
+
 def urlencoded_params_matcher(params):
     warn(
         "Function is deprecated. Use 'from responses.matchers import urlencoded_params_matcher'",
@@ -87,16 +78,26 @@ def json_params_matcher(params):
     return _json_params_matcher(params)
 
 
-def _is_string(s):
-    return isinstance(s, six.string_types)
-
-
 def _has_unicode(s):
     return any(ord(char) > 128 for char in s)
 
 
 def _clean_unicode(url):
-    # Clean up domain names, which use punycode to handle unicode chars
+    """Clean up URLs, which use punycode to handle unicode chars.
+
+    Applies percent encoding to URL path and query if required.
+
+    Parameters
+    ----------
+    url : str
+        URL that should be cleaned from unicode
+
+    Returns
+    -------
+    str
+        Cleaned URL
+
+    """
     urllist = list(urlsplit(url))
     netloc = urllist[1]
     if _has_unicode(netloc):
@@ -109,8 +110,6 @@ def _clean_unicode(url):
         url = urlunsplit(urllist)
 
     # Clean up path/query/params, which use url-encoding to handle unicode chars
-    if isinstance(url.encode("utf8"), six.string_types):
-        url = url.encode("utf8")
     chars = list(url)
     for i, x in enumerate(chars):
         if ord(x) > 128:
@@ -119,97 +118,67 @@ def _clean_unicode(url):
     return "".join(chars)
 
 
-def _is_redirect(response):
-    try:
-        # 2.0.0 <= requests <= 2.2
-        return response.is_redirect
-
-    except AttributeError:
-        # requests > 2.2
-        return (
-            # use request.sessions conditional
-            response.status_code in REDIRECT_STATI
-            and "location" in response.headers
-        )
-
-
-def _ensure_str(s):
-    if six.PY2:
-        s = s.encode("utf-8") if isinstance(s, six.text_type) else s
-    return s
-
-
 def _cookies_from_headers(headers):
-    try:
-        import http.cookies as cookies
+    """Create Cookies from request Headers.
 
-        resp_cookie = cookies.SimpleCookie()
-        resp_cookie.load(headers["set-cookie"])
+    Converts ``set-cookie`` headers to real cookies.
 
-        cookies_dict = {name: v.value for name, v in resp_cookie.items()}
-    except ImportError:
-        from cookies import Cookies
+    Parameters
+    ----------
+    headers : dict
+        Request headers.
+
+    Returns
+    -------
+    CookieJar
+        CookieJar request object.
+
+    """
+    resp_cookie = cookies.SimpleCookie()
+    resp_cookie.load(headers["set-cookie"])
+    cookies_dict = {name: v.value for name, v in resp_cookie.items()}
 
-        resp_cookies = Cookies.from_request(_ensure_str(headers["set-cookie"]))
-        cookies_dict = {
-            v.name: quote(_ensure_str(v.value)) for _, v in resp_cookies.items()
-        }
     return cookiejar_from_dict(cookies_dict)
 
 
-_wrapper_template = """\
-def wrapper%(wrapper_args)s:
-    with responses:
-        return func%(func_args)s
-"""
+def get_wrapped(func, responses, registry=None):
+    """Wrap provided function inside ``responses`` context manager.
+
+    Provides a synchronous or asynchronous wrapper for the function.
 
 
-def get_wrapped(func, responses):
-    if six.PY2:
-        args, a, kw, defaults = inspect.getargspec(func)
-        wrapper_args = inspect.formatargspec(args, a, kw, defaults)
+    Parameters
+    ----------
+    func : Callable
+        Function to wrap.
+    responses : RequestsMock
+        Mock object that is used as context manager.
+    registry : FirstMatchRegistry, optional
+        Custom registry that should be applied. See ``responses.registries``
+
+    Returns
+    -------
+    Callable
+        Wrapped function
+
+    """
+    if registry is not None:
+        responses._set_registry(registry)
+
+    if inspect.iscoroutinefunction(func):
+        # set asynchronous wrapper if requestor function is asynchronous
+        @wraps(func)
+        async def wrapper(*args, **kwargs):
+            with responses:
+                return await func(*args, **kwargs)
 
-        # Preserve the argspec for the wrapped function so that testing
-        # tools such as pytest can continue to use their fixture injection.
-        if hasattr(func, "__self__"):
-            args = args[1:]  # Omit 'self'
-        func_args = inspect.formatargspec(args, a, kw, None)
     else:
-        signature = inspect.signature(func)
-        signature = signature.replace(return_annotation=inspect.Signature.empty)
-        # If the function is wrapped, switch to *args, **kwargs for the parameters
-        # as we can't rely on the signature to give us the arguments the function will
-        # be called with. For example unittest.mock.patch uses required args that are
-        # not actually passed to the function when invoked.
-        if hasattr(func, "__wrapped__"):
-            wrapper_params = [
-                inspect.Parameter("args", inspect.Parameter.VAR_POSITIONAL),
-                inspect.Parameter("kwargs", inspect.Parameter.VAR_KEYWORD),
-            ]
-        else:
-            wrapper_params = [
-                param.replace(annotation=inspect.Parameter.empty)
-                for param in signature.parameters.values()
-            ]
-        signature = signature.replace(parameters=wrapper_params)
 
-        wrapper_args = str(signature)
-        params_without_defaults = [
-            param.replace(
-                annotation=inspect.Parameter.empty, default=inspect.Parameter.empty
-            )
-            for param in signature.parameters.values()
-        ]
-        signature = signature.replace(parameters=params_without_defaults)
-        func_args = str(signature)
+        @wraps(func)
+        def wrapper(*args, **kwargs):
+            with responses:
+                return func(*args, **kwargs)
 
-    evaldict = {"func": func, "responses": responses}
-    six.exec_(
-        _wrapper_template % {"wrapper_args": wrapper_args, "func_args": func_args},
-        evaldict,
-    )
-    wrapper = evaldict["wrapper"]
-    update_wrapper(wrapper, func)
     return wrapper
 
 
@@ -234,7 +203,7 @@ class CallList(Sequence, Sized):
 
 
 def _ensure_url_default_path(url):
-    if _is_string(url):
+    if isinstance(url, str):
         url_parts = list(urlsplit(url))
         if url_parts[2] == "":
             url_parts[2] = "/"
@@ -242,29 +211,65 @@ def _ensure_url_default_path(url):
     return url
 
 
+def _get_url_and_path(url):
+    url_parsed = urlsplit(url)
+    url_and_path = urlunparse(
+        [url_parsed.scheme, url_parsed.netloc, url_parsed.path, None, None, None]
+    )
+    return parse_url(url_and_path).url
+
+
 def _handle_body(body):
-    if isinstance(body, six.text_type):
+    if isinstance(body, str):
         body = body.encode("utf-8")
-    if isinstance(body, _io.BufferedReader):
+    if isinstance(body, BufferedReader):
         return body
 
-    return BufferIO(body)
+    data = BytesIO(body)
 
+    def is_closed():
+        """
+        Real Response uses HTTPResponse as body object.
+        Thus, when method is_closed is called first to check if there is any more
+        content to consume and the file-like object is still opened
 
-_unspecified = object()
+        This method ensures stability to work for both:
+        https://github.com/getsentry/responses/issues/438
+        https://github.com/getsentry/responses/issues/394
+
+        where file should be intentionally be left opened to continue consumption
+        """
+        if not data.closed and data.read(1):
+            # if there is more bytes to read then keep open, but return pointer
+            data.seek(-1, 1)
+            return False
+        else:
+            if not data.closed:
+                # close but return False to mock like is still opened
+                data.close()
+                return False
+
+            # only if file really closed (by us) return True
+            return True
+
+    data.isclosed = is_closed
+    return data
 
 
 class BaseResponse(object):
+    passthrough = False
     content_type = None
     headers = None
-
     stream = False
 
-    def __init__(self, method, url, match_querystring=_unspecified, match=[]):
+    def __init__(self, method, url, match_querystring=None, match=()):
         self.method = method
         # ensure the url has a default path set if the url is a string
         self.url = _ensure_url_default_path(url)
-        self.match_querystring = self._should_match_querystring(match_querystring)
+
+        if self._should_match_querystring(match_querystring):
+            match = tuple(match) + (_query_string_matcher(urlsplit(self.url).query),)
+
         self.match = match
         self.call_count = 0
 
@@ -286,45 +291,31 @@ class BaseResponse(object):
     def __ne__(self, other):
         return not self.__eq__(other)
 
-    def _url_matches_strict(self, url, other):
-        url_parsed = urlparse(url)
-        other_parsed = urlparse(other)
-
-        if url_parsed[:3] != other_parsed[:3]:
-            return False
-
-        url_qsl = sorted(parse_qsl(url_parsed.query))
-        other_qsl = sorted(parse_qsl(other_parsed.query))
-
-        return url_qsl == other_qsl
-
     def _should_match_querystring(self, match_querystring_argument):
-        if match_querystring_argument is not _unspecified:
-            return match_querystring_argument
-
         if isinstance(self.url, Pattern):
             # the old default from <= 0.9.0
             return False
 
-        return bool(urlparse(self.url).query)
+        if match_querystring_argument is not None:
+            if not isinstance(match_querystring_argument, FalseBool):
+                warn(
+                    (
+                        "Argument 'match_querystring' is deprecated. "
+                        "Use 'responses.matchers.query_param_matcher' or "
+                        "'responses.matchers.query_string_matcher'"
+                    ),
+                    DeprecationWarning,
+                )
+            return match_querystring_argument
 
-    def _url_matches(self, url, other, match_querystring=False):
-        if _is_string(url):
+        return bool(urlsplit(self.url).query)
+
+    def _url_matches(self, url, other):
+        if isinstance(url, str):
             if _has_unicode(url):
                 url = _clean_unicode(url)
-                if not isinstance(other, six.text_type):
-                    other = other.encode("ascii").decode("utf8")
 
-            if match_querystring:
-                normalize_url = parse_url(url).url
-                return self._url_matches_strict(normalize_url, other)
-
-            else:
-                url_without_qs = url.split("?", 1)[0]
-                other_without_qs = other.split("?", 1)[0]
-                normalized_url_without_qs = parse_url(url_without_qs).url
-
-                return normalized_url_without_qs == other_without_qs
+            return _get_url_and_path(url) == _get_url_and_path(other)
 
         elif isinstance(url, Pattern) and url.match(other):
             return True
@@ -356,12 +347,12 @@ class BaseResponse(object):
         if request.method != self.method:
             return False, "Method does not match"
 
-        if not self._url_matches(self.url, request.url, self.match_querystring):
+        if not self._url_matches(self.url, request.url):
             return False, "URL does not match"
 
         valid, reason = self._req_attr_matches(self.match, request)
         if not valid:
-            return False, "Parameters do not match. " + reason
+            return False, reason
 
         return True, ""
 
@@ -375,21 +366,21 @@ class Response(BaseResponse):
         json=None,
         status=200,
         headers=None,
-        stream=False,
-        content_type=UNSET,
+        stream=None,
+        content_type=_UNSET,
         auto_calculate_content_length=False,
-        **kwargs
+        **kwargs,
     ):
         # if we were passed a `json` argument,
         # override the body and content_type
         if json is not None:
             assert not body
             body = json_module.dumps(json)
-            if content_type is UNSET:
+            if content_type is _UNSET:
                 content_type = "application/json"
 
-        if content_type is UNSET:
-            if isinstance(body, six.text_type) and _has_unicode(body):
+        if content_type is _UNSET:
+            if isinstance(body, str) and _has_unicode(body):
                 content_type = "text/plain; charset=utf-8"
             else:
                 content_type = "text/plain"
@@ -397,10 +388,17 @@ class Response(BaseResponse):
         self.body = body
         self.status = status
         self.headers = headers
+
+        if stream is not None:
+            warn(
+                "stream argument is deprecated. Use stream parameter in request directly",
+                DeprecationWarning,
+            )
+
         self.stream = stream
         self.content_type = content_type
         self.auto_calculate_content_length = auto_calculate_content_length
-        super(Response, self).__init__(method, url, **kwargs)
+        super().__init__(method, url, **kwargs)
 
     def get_response(self, request):
         if self.body and isinstance(self.body, Exception):
@@ -412,7 +410,7 @@ class Response(BaseResponse):
 
         if (
             self.auto_calculate_content_length
-            and isinstance(body, BufferIO)
+            and isinstance(body, BytesIO)
             and "Content-Length" not in headers
         ):
             content_length = len(body.getvalue())
@@ -420,7 +418,7 @@ class Response(BaseResponse):
 
         return HTTPResponse(
             status=status,
-            reason=six.moves.http_client.responses.get(status),
+            reason=client.responses.get(status, None),
             body=body,
             headers=headers,
             original_response=OriginalResponseShim(headers),
@@ -441,12 +439,18 @@ class Response(BaseResponse):
 
 class CallbackResponse(BaseResponse):
     def __init__(
-        self, method, url, callback, stream=False, content_type="text/plain", **kwargs
+        self, method, url, callback, stream=None, content_type="text/plain", **kwargs
     ):
         self.callback = callback
+
+        if stream is not None:
+            warn(
+                "stream argument is deprecated. Use stream parameter in request directly",
+                DeprecationWarning,
+            )
         self.stream = stream
         self.content_type = content_type
-        super(CallbackResponse, self).__init__(method, url, **kwargs)
+        super().__init__(method, url, **kwargs)
 
     def get_response(self, request):
         headers = self.get_headers()
@@ -477,7 +481,7 @@ class CallbackResponse(BaseResponse):
 
         return HTTPResponse(
             status=status,
-            reason=six.moves.http_client.responses.get(status),
+            reason=client.responses.get(status, None),
             body=body,
             headers=headers,
             original_response=OriginalResponseShim(headers),
@@ -485,6 +489,10 @@ class CallbackResponse(BaseResponse):
         )
 
 
+class PassthroughResponse(BaseResponse):
+    passthrough = True
+
+
 class OriginalResponseShim(object):
     """
     Shim for compatibility with older versions of urllib3
@@ -504,6 +512,9 @@ class OriginalResponseShim(object):
     def isclosed(self):
         return True
 
+    def close(self):
+        return
+
 
 class RequestsMock(object):
     DELETE = "DELETE"
@@ -521,17 +532,35 @@ class RequestsMock(object):
         response_callback=None,
         passthru_prefixes=(),
         target="requests.adapters.HTTPAdapter.send",
+        registry=FirstMatchRegistry,
     ):
         self._calls = CallList()
         self.reset()
+        self._registry = registry()  # call only after reset
         self.assert_all_requests_are_fired = assert_all_requests_are_fired
         self.response_callback = response_callback
         self.passthru_prefixes = tuple(passthru_prefixes)
         self.target = target
+        self._patcher = None
+
+    def get_registry(self):
+        return self._registry
+
+    def _set_registry(self, new_registry):
+        if self.registered():
+            err_msg = (
+                "Cannot replace Registry, current registry has responses.\n"
+                "Run 'responses.registry.reset()' first"
+            )
+            raise AttributeError(err_msg)
+
+        self._registry = new_registry()
 
     def reset(self):
-        self._matches = []
+        self._registry = FirstMatchRegistry()
         self._calls.reset()
+        self.passthru_prefixes = ()
+        self._patcher = None
 
     def add(
         self,
@@ -540,11 +569,12 @@ class RequestsMock(object):
         body="",
         adding_headers=None,
         *args,
-        **kwargs
+        **kwargs,
     ):
         """
-        A basic request:
+        >>> import responses
 
+        A basic request:
         >>> responses.add(responses.GET, 'http://example.com')
 
         You can also directly pass an object which implements the
@@ -568,23 +598,43 @@ class RequestsMock(object):
         >>>     headers={'X-Header': 'foo'},
         >>> )
 
-
-        Strict query string matching:
-
-        >>> responses.add(
-        >>>     method='GET',
-        >>>     url='http://example.com?foo=bar',
-        >>>     match_querystring=True
-        >>> )
         """
         if isinstance(method, BaseResponse):
-            self._matches.append(method)
+            self._registry.add(method)
             return
 
         if adding_headers is not None:
             kwargs.setdefault("headers", adding_headers)
+        if "content_type" in kwargs and "headers" in kwargs:
+            header_keys = [header.lower() for header in kwargs["headers"]]
+            if "content-type" in header_keys:
+                raise RuntimeError(
+                    "You cannot define both `content_type` and `headers[Content-Type]`."
+                    " Using the `content_type` kwarg is recommended."
+                )
 
-        self._matches.append(Response(method=method, url=url, body=body, **kwargs))
+        self._registry.add(Response(method=method, url=url, body=body, **kwargs))
+
+    def delete(self, *args, **kwargs):
+        self.add(DELETE, *args, **kwargs)
+
+    def get(self, *args, **kwargs):
+        self.add(GET, *args, **kwargs)
+
+    def head(self, *args, **kwargs):
+        self.add(HEAD, *args, **kwargs)
+
+    def options(self, *args, **kwargs):
+        self.add(OPTIONS, *args, **kwargs)
+
+    def patch(self, *args, **kwargs):
+        self.add(PATCH, *args, **kwargs)
+
+    def post(self, *args, **kwargs):
+        self.add(POST, *args, **kwargs)
+
+    def put(self, *args, **kwargs):
+        self.add(PUT, *args, **kwargs)
 
     def add_passthru(self, prefix):
         """
@@ -593,10 +643,12 @@ class RequestsMock(object):
         For example, to allow any request to 'https://example.com', but require
         mocks for the remainder, you would add the prefix as so:
 
+        >>> import responses
         >>> responses.add_passthru('https://example.com')
 
         Regex can be used like:
 
+        >>> import re
         >>> responses.add_passthru(re.compile('https://example.com/\\w+'))
         """
         if not isinstance(prefix, Pattern) and _has_unicode(prefix):
@@ -609,16 +661,16 @@ class RequestsMock(object):
         either by a response object inheriting ``BaseResponse`` or
         ``method`` and ``url``. Removes all matching responses.
 
-        >>> response.add(responses.GET, 'http://example.org')
-        >>> response.remove(responses.GET, 'http://example.org')
+        >>> import responses
+        >>> responses.add(responses.GET, 'http://example.org')
+        >>> responses.remove(responses.GET, 'http://example.org')
         """
         if isinstance(method_or_response, BaseResponse):
             response = method_or_response
         else:
             response = BaseResponse(method=method_or_response, url=url)
 
-        while response in self._matches:
-            self._matches.remove(response)
+        self._registry.remove(response)
 
     def replace(self, method_or_response=None, url=None, body="", *args, **kwargs):
         """
@@ -626,20 +678,16 @@ class RequestsMock(object):
         is identical to ``add()``. The response is identified using ``method``
         and ``url``, and the first matching response is replaced.
 
+        >>> import responses
         >>> responses.add(responses.GET, 'http://example.org', json={'data': 1})
         >>> responses.replace(responses.GET, 'http://example.org', json={'data': 2})
         """
         if isinstance(method_or_response, BaseResponse):
-            url = method_or_response.url
             response = method_or_response
         else:
             response = Response(method=method_or_response, url=url, body=body, **kwargs)
 
-        try:
-            index = self._matches.index(response)
-        except ValueError:
-            raise ValueError("Response is not registered for URL %s" % url)
-        self._matches[index] = response
+        self._registry.replace(response)
 
     def upsert(self, method_or_response=None, url=None, body="", *args, **kwargs):
         """
@@ -647,6 +695,7 @@ class RequestsMock(object):
         if no response exists.  Responses are matched using ``method``and ``url``.
         The first matching response is replaced.
 
+        >>> import responses
         >>> responses.add(responses.GET, 'http://example.org', json={'data': 1})
         >>> responses.upsert(responses.GET, 'http://example.org', json={'data': 2})
         """
@@ -656,23 +705,28 @@ class RequestsMock(object):
             self.add(method_or_response, url, body, *args, **kwargs)
 
     def add_callback(
-        self, method, url, callback, match_querystring=False, content_type="text/plain"
+        self,
+        method,
+        url,
+        callback,
+        match_querystring=FalseBool(),
+        content_type="text/plain",
+        match=(),
     ):
-        # ensure the url has a default path set if the url is a string
-        # url = _ensure_url_default_path(url, match_querystring)
 
-        self._matches.append(
+        self._registry.add(
             CallbackResponse(
                 url=url,
                 method=method,
                 callback=callback,
                 content_type=content_type,
                 match_querystring=match_querystring,
+                match=match,
             )
         )
 
     def registered(self):
-        return self._matches
+        return self._registry.registered
 
     @property
     def calls(self):
@@ -688,29 +742,29 @@ class RequestsMock(object):
         self.reset()
         return success
 
-    def activate(self, func):
-        return get_wrapped(func, self)
+    def activate(self, func=None, registry=None):
+        if func is not None:
+            return get_wrapped(func, self)
+
+        def deco_activate(function):
+            return get_wrapped(function, self, registry)
+
+        return deco_activate
 
     def _find_match(self, request):
-        found = None
-        found_match = None
-        match_failed_reasons = []
-        for i, match in enumerate(self._matches):
-            match_result, reason = match.matches(request)
-            if match_result:
-                if found is None:
-                    found = i
-                    found_match = match
-                else:
-                    # Multiple matches found.  Remove & return the first match.
-                    return self._matches.pop(found), match_failed_reasons
-            else:
-                match_failed_reasons.append(reason)
-        return found_match, match_failed_reasons
+        """
+        Iterates through all available matches and validates if any of them matches the request
+
+        :param request: (PreparedRequest), request object
+        :return:
+            (Response) found match. If multiple found, then remove & return the first match.
+            (list) list with reasons why other matches don't match
+        """
+        return self._registry.find(request)
 
     def _parse_request_params(self, url):
         params = {}
-        for key, val in groupby(parse_qsl(urlparse(url).query), lambda kv: kv[0]):
+        for key, val in groupby(parse_qsl(urlsplit(url).query), lambda kv: kv[0]):
             values = list(map(lambda x: x[1], val))
             if len(values) == 1:
                 values = values[0]
@@ -718,7 +772,11 @@ class RequestsMock(object):
         return params
 
     def _on_request(self, adapter, request, **kwargs):
+        # add attributes params and req_kwargs to 'request' object for further match comparison
+        # original request object does not have these attributes
         request.params = self._parse_request_params(request.path_url)
+        request.req_kwargs = kwargs
+
         match, match_failed_reasons = self._find_match(request)
         resp_callback = self.response_callback
 
@@ -738,31 +796,35 @@ class RequestsMock(object):
                 "Connection refused by Responses - the call doesn't "
                 "match any registered mock.\n\n"
                 "Request: \n"
-                "- %s %s\n\n"
-                "Available matches:\n" % (request.method, request.url)
+                f"- {request.method} {request.url}\n\n"
+                "Available matches:\n"
             )
-            for i, m in enumerate(self._matches):
+            for i, m in enumerate(self.registered()):
                 error_msg += "- {} {} {}\n".format(
                     m.method, m.url, match_failed_reasons[i]
                 )
 
+            if self.passthru_prefixes:
+                error_msg += "Passthru prefixes:\n"
+                for p in self.passthru_prefixes:
+                    error_msg += "- {}\n".format(p)
+
             response = ConnectionError(error_msg)
             response.request = request
 
             self._calls.add(request, response)
-            response = resp_callback(response) if resp_callback else response
             raise response
 
-        try:
-            response = adapter.build_response(request, match.get_response(request))
-        except BaseException as response:
-            match.call_count += 1
-            self._calls.add(request, response)
-            response = resp_callback(response) if resp_callback else response
-            raise
-
-        if not match.stream:
-            response.content  # NOQA
+        if match.passthrough:
+            logger.info("request.passthrough-response", extra={"url": request.url})
+            response = _real_send(adapter, request, **kwargs)
+        else:
+            try:
+                response = adapter.build_response(request, match.get_response(request))
+            except BaseException as response:
+                match.call_count += 1
+                self._calls.add(request, response)
+                raise
 
         response = resp_callback(response) if resp_callback else response
         match.call_count += 1
@@ -770,6 +832,10 @@ class RequestsMock(object):
         return response
 
     def start(self):
+        if self._patcher:
+            # we must not override value of the _patcher if already applied
+            return
+
         def unbound_on_send(adapter, request, *a, **kwargs):
             return self._on_request(adapter, request, *a, **kwargs)
 
@@ -777,14 +843,17 @@ class RequestsMock(object):
         self._patcher.start()
 
     def stop(self, allow_assert=True):
-        self._patcher.stop()
+        if self._patcher:
+            # prevent stopping unstarted patchers
+            self._patcher.stop()
+
         if not self.assert_all_requests_are_fired:
             return
 
         if not allow_assert:
             return
 
-        not_called = [m for m in self._matches if m.call_count == 0]
+        not_called = [m for m in self.registered() if m.call_count == 0]
         if not_called:
             raise AssertionError(
                 "Not all requests have been executed {0!r}".format(
@@ -804,9 +873,7 @@ class RequestsMock(object):
             return True
         else:
             raise AssertionError(
-                "Expected URL '{0}' to be called {1} times. Called {2} times.".format(
-                    url, count, call_count
-                )
+                f"Expected URL '{url}' to be called {count} times. Called {call_count} times."
             )
 
 
@@ -824,13 +891,20 @@ __all__ = [
     "assert_all_requests_are_fired",
     "assert_call_count",
     "calls",
+    "delete",
     "DELETE",
+    "get",
     "GET",
+    "head",
     "HEAD",
+    "options",
     "OPTIONS",
     "passthru_prefixes",
+    "patch",
     "PATCH",
+    "post",
     "POST",
+    "put",
     "PUT",
     "registered",
     "remove",
@@ -850,13 +924,20 @@ add_passthru = _default_mock.add_passthru
 assert_all_requests_are_fired = _default_mock.assert_all_requests_are_fired
 assert_call_count = _default_mock.assert_call_count
 calls = _default_mock.calls
+delete = _default_mock.delete
 DELETE = _default_mock.DELETE
+get = _default_mock.get
 GET = _default_mock.GET
+head = _default_mock.head
 HEAD = _default_mock.HEAD
+options = _default_mock.options
 OPTIONS = _default_mock.OPTIONS
 passthru_prefixes = _default_mock.passthru_prefixes
+patch = _default_mock.patch
 PATCH = _default_mock.PATCH
+post = _default_mock.post
 POST = _default_mock.POST
+put = _default_mock.put
 PUT = _default_mock.PUT
 registered = _default_mock.registered
 remove = _default_mock.remove
