@@ -1,34 +1,30 @@
 # coding: utf-8
 
-from __future__ import absolute_import, print_function, division, unicode_literals
-
 import inspect
 import os
 import re
-import six
-from io import BufferedReader, BytesIO
+import warnings
+from io import BufferedReader
+from io import BytesIO
+from unittest.mock import Mock
+from unittest.mock import patch
 
 import pytest
 import requests
+from requests.exceptions import ChunkedEncodingError
+from requests.exceptions import ConnectionError
+from requests.exceptions import HTTPError
+
 import responses
-from requests.exceptions import ConnectionError, HTTPError, ChunkedEncodingError
-from responses import (
-    BaseResponse,
-    Response,
-    PassthroughResponse,
-    matchers,
-    CallbackResponse,
-)
-
-
-try:
-    from mock import patch, Mock
-except ImportError:
-    from unittest.mock import patch, Mock  # type: ignore
+from responses import BaseResponse
+from responses import CallbackResponse
+from responses import PassthroughResponse
+from responses import Response
+from responses import matchers
 
 
 def assert_reset():
-    assert len(responses._default_mock._matches) == 0
+    assert len(responses._default_mock.registered()) == 0
     assert len(responses.calls) == 0
 
 
@@ -559,11 +555,46 @@ def test_callback():
     assert_reset()
 
 
-def test_callback_deprecated_argument():
+def test_callback_deprecated_stream_argument():
     with pytest.deprecated_call():
         CallbackResponse(responses.GET, "url", lambda x: x, stream=False)
 
 
+def test_callback_deprecated_match_querystring_argument():
+    with pytest.deprecated_call():
+        CallbackResponse(responses.GET, "url", lambda x: x, match_querystring=False)
+
+
+def test_callback_match_querystring_default_false():
+    """
+    Test to ensure that by default 'match_querystring' in 'add_callback' is set to False
+    and does not raise deprecation
+    see: https://github.com/getsentry/responses/issues/464 and related PR
+    """
+    body = b"test callback"
+    status = 200
+    params = {"hello": "world", "I am": "a big test"}
+    headers = {"foo": "bar"}
+    url = "http://example.com/"
+
+    def request_callback(_request):
+        return status, headers, body
+
+    @responses.activate
+    def run():
+        responses.add_callback(responses.GET, url, request_callback, content_type=None)
+        resp = requests.get(url, params=params)
+        assert resp.text == "test callback"
+        assert resp.status_code == status
+        assert "foo" in resp.headers
+
+    with warnings.catch_warnings():
+        warnings.simplefilter("error")
+        run()
+
+    assert_reset()
+
+
 def test_callback_exception_result():
     result = Exception()
     url = "http://example.com/"
@@ -761,7 +792,7 @@ def test_custom_adapter():
         class DummyAdapter(requests.adapters.HTTPAdapter):
             def send(self, *a, **k):
                 calls[0] += 1
-                return super(DummyAdapter, self).send(*a, **k)
+                return super().send(*a, **k)
 
         # Test that the adapter is actually used
         session = requests.Session()
@@ -805,14 +836,10 @@ def test_activate_doesnt_change_signature():
         return (a, b)
 
     decorated_test_function = responses.activate(test_function)
-    if hasattr(inspect, "signature"):
-        assert inspect.signature(test_function) == inspect.signature(
-            decorated_test_function
-        )
-    else:
-        assert inspect.getargspec(test_function) == inspect.getargspec(
-            decorated_test_function
-        )
+    assert inspect.signature(test_function) == inspect.signature(
+        decorated_test_function
+    )
+
     assert decorated_test_function(1, 2) == test_function(1, 2)
     assert decorated_test_function(3) == test_function(3)
 
@@ -846,14 +873,9 @@ def test_activate_mock_interaction():
         return mock_stdout
 
     decorated_test_function = responses.activate(test_function)
-    if hasattr(inspect, "signature"):
-        assert inspect.signature(test_function) == inspect.signature(
-            decorated_test_function
-        )
-    else:
-        assert inspect.getargspec(test_function) == inspect.getargspec(
-            decorated_test_function
-        )
+    assert inspect.signature(test_function) == inspect.signature(
+        decorated_test_function
+    )
 
     value = test_function()
     assert isinstance(value, Mock)
@@ -862,7 +884,6 @@ def test_activate_mock_interaction():
     assert isinstance(value, Mock)
 
 
-@pytest.mark.skipif(six.PY2, reason="Cannot run in python2")
 def test_activate_doesnt_change_signature_with_return_type():
     def test_function(a, b=None):
         return a, b
@@ -1016,7 +1037,6 @@ def test_response_callback():
     assert_reset()
 
 
-@pytest.mark.skipif(six.PY2, reason="re.compile works differntly in PY2")
 def test_response_filebody():
     """ Adds the possibility to use actual (binary) files as responses """
 
@@ -1070,24 +1090,24 @@ def test_assert_all_requests_are_fired():
         # check that assert_all_requests_are_fired=True doesn't remove urls
         with responses.RequestsMock(assert_all_requests_are_fired=True) as m:
             m.add(responses.GET, "http://example.com", body=b"test")
-            assert len(m._matches) == 1
+            assert len(m.registered()) == 1
             requests.get("http://example.com")
-            assert len(m._matches) == 1
+            assert len(m.registered()) == 1
 
         # check that assert_all_requests_are_fired=True counts mocked errors
         with responses.RequestsMock(assert_all_requests_are_fired=True) as m:
             m.add(responses.GET, "http://example.com", body=Exception())
-            assert len(m._matches) == 1
+            assert len(m.registered()) == 1
             with pytest.raises(Exception):
                 requests.get("http://example.com")
-            assert len(m._matches) == 1
+            assert len(m.registered()) == 1
 
         with responses.RequestsMock(assert_all_requests_are_fired=True) as m:
             m.add_callback(responses.GET, "http://example.com", request_callback)
-            assert len(m._matches) == 1
+            assert len(m.registered()) == 1
             with pytest.raises(BaseException):
                 requests.get("http://example.com")
-            assert len(m._matches) == 1
+            assert len(m.registered()) == 1
 
     run()
     assert_reset()
@@ -1138,6 +1158,28 @@ def test_allow_redirects_samehost():
     assert_reset()
 
 
+def test_path_segments():
+    """Test that path segment after ``;`` is preserved.
+
+    Validate compliance with RFC 3986.
+    The path is terminated by the first question mark ("?") or
+    number sign ("#") character, or by the end of the URI.
+    See more about how path should be treated under:
+    https://datatracker.ietf.org/doc/html/rfc3986.html#section-3.3
+    """
+
+    @responses.activate
+    def run():
+        responses.add(responses.GET, "http://example.com/here/we", status=669)
+        responses.add(responses.GET, "http://example.com/here/we;go", status=777)
+
+        resp = requests.get("http://example.com/here/we;go")
+        assert resp.status_code == 777
+
+    run()
+    assert_reset()
+
+
 def test_handles_unicode_querystring():
     url = "http://example.com/test?type=2&ie=utf8&query=汉字"
 
@@ -1219,6 +1261,9 @@ def test_content_length_error(monkeypatch):
     https://github.com/psf/requests/pull/3563
 
     Now user can manually patch URL3 lib to achieve the same
+
+    See discussion in
+    https://github.com/getsentry/responses/issues/394
     """
 
     @responses.activate
@@ -1234,20 +1279,47 @@ def test_content_length_error(monkeypatch):
 
         assert "IncompleteRead" in str(exc.value)
 
-    original_init = getattr(requests.packages.urllib3.HTTPResponse, "__init__")
+    # Type errors here and on 1250 are ignored because the stubs for requests
+    # are off https://github.com/python/typeshed/blob/f8501d33c737482a829c6db557a0be26895c5941
+    #   /stubs/requests/requests/packages/__init__.pyi#L1
+    original_init = getattr(requests.packages.urllib3.HTTPResponse, "__init__")  # type: ignore
 
     def patched_init(self, *args, **kwargs):
         kwargs["enforce_content_length"] = True
         original_init(self, *args, **kwargs)
 
     monkeypatch.setattr(
-        requests.packages.urllib3.HTTPResponse, "__init__", patched_init
+        requests.packages.urllib3.HTTPResponse, "__init__", patched_init  # type: ignore
     )
 
     run()
     assert_reset()
 
 
+def test_stream_with_none_chunk_size():
+    """
+    See discussion in
+    https://github.com/getsentry/responses/issues/438
+    """
+
+    @responses.activate
+    def run():
+        responses.add(
+            responses.GET,
+            "https://example.com",
+            status=200,
+            content_type="application/octet-stream",
+            body=b"This is test",
+            auto_calculate_content_length=True,
+        )
+        res = requests.get("https://example.com", stream=True)
+        for chunk in res.iter_content(chunk_size=None):
+            assert chunk == b"This is test"
+
+    run()
+    assert_reset()
+
+
 def test_legacy_adding_headers():
     @responses.activate
     def run():
@@ -1264,6 +1336,26 @@ def test_legacy_adding_headers():
     assert_reset()
 
 
+def test_legacy_adding_headers_with_content_type():
+    @responses.activate
+    def run():
+        with pytest.raises(RuntimeError) as excinfo:
+            responses.add(
+                responses.GET,
+                "http://example.com",
+                body="test",
+                content_type="text/html",
+                adding_headers={"Content-Type": "text/html; charset=utf-8"},
+            )
+        assert (
+            "You cannot define both `content_type` and `headers[Content-Type]`"
+            in str(excinfo.value)
+        )
+
+    run()
+    assert_reset()
+
+
 def test_auto_calculate_content_length_string_body():
     @responses.activate
     def run():
@@ -1377,15 +1469,48 @@ def test_multiple_responses():
     def run():
         responses.add(responses.GET, "http://example.com", body="test")
         responses.add(responses.GET, "http://example.com", body="rest")
+        responses.add(responses.GET, "http://example.com", body="fest")
+        responses.add(responses.GET, "http://example.com", body="best")
 
         resp = requests.get("http://example.com")
         assert_response(resp, "test")
+
         resp = requests.get("http://example.com")
         assert_response(resp, "rest")
+
+        resp = requests.get("http://example.com")
+        assert_response(resp, "fest")
+
+        resp = requests.get("http://example.com")
+        assert_response(resp, "best")
+
         # After all responses are used, last response should be repeated
         resp = requests.get("http://example.com")
+        assert_response(resp, "best")
+
+    run()
+    assert_reset()
+
+
+def test_multiple_responses_intermixed():
+    @responses.activate
+    def run():
+        responses.add(responses.GET, "http://example.com", body="test")
+        resp = requests.get("http://example.com")
+        assert_response(resp, "test")
+
+        responses.add(responses.GET, "http://example.com", body="rest")
+        resp = requests.get("http://example.com")
         assert_response(resp, "rest")
 
+        responses.add(responses.GET, "http://example.com", body="best")
+        resp = requests.get("http://example.com")
+        assert_response(resp, "best")
+
+        # After all responses are used, last response should be repeated
+        resp = requests.get("http://example.com")
+        assert_response(resp, "best")
+
     run()
     assert_reset()
 
@@ -1546,6 +1671,37 @@ def test_passthru_regex(httpserver):
     assert_reset()
 
 
+def test_passthru_does_not_persist_across_tests(httpserver):
+    """
+    passthru should be erased on exit from context manager
+    see:
+    https://github.com/getsentry/responses/issues/322
+    """
+    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
+
+    @responses.activate
+    def with_a_passthru():
+        assert not responses._default_mock.passthru_prefixes
+        responses.add_passthru(re.compile(".*"))
+
+        # wrap request that is passed through with another mock. That helps
+        # to avoid issues if real URL is unavailable, allow to run tests offline
+        with responses.RequestsMock(target="responses._real_send") as rsp:
+            rsp.add(responses.GET, "https://example66.ru", status=969)
+            response = requests.get("https://example66.ru")
+
+            assert response.status_code == 969
+
+    @responses.activate
+    def without_a_passthru():
+        assert not responses._default_mock.passthru_prefixes
+        with pytest.raises(requests.exceptions.ConnectionError):
+            requests.get("https://example.com")
+
+    with_a_passthru()
+    without_a_passthru()
+
+
 def test_method_named_param():
     @responses.activate
     def run():
@@ -1580,13 +1736,9 @@ def test_custom_target(monkeypatch):
     assert patch_mock.call_args[1]["target"] == "something.else"
 
 
-def _quote(s):
-    return responses.quote(responses._ensure_str(s))
-
-
 def test_cookies_from_headers():
     text = "こんにちは/世界"
-    quoted_text = _quote(text)
+    quoted_text = responses.quote(text)
     expected = {"x": "a", "y": quoted_text}
     headers = {"set-cookie": "; ".join(k + "=" + v for k, v in expected.items())}
     cookiejar = responses._cookies_from_headers(headers)
@@ -1687,6 +1839,7 @@ def test_fail_request_error():
         with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
             rsps.add("POST", "http://example1.com")
             rsps.add("GET", "http://example.com")
+            rsps.add_passthru("http://other.example.com")
 
             with pytest.raises(ConnectionError) as excinfo:
                 requests.post("http://example.com", data={"id": "bad"})
@@ -1694,6 +1847,7 @@ def test_fail_request_error():
             msg = str(excinfo.value)
             assert "- POST http://example1.com/ URL does not match" in msg
             assert "- GET http://example.com/ Method does not match" in msg
+            assert "Passthru prefixes:\n- http://other.example.com" in msg
 
     run()
     assert_reset()
@@ -1767,7 +1921,7 @@ def test_mocked_responses_list_registered():
 
         mocks_list = responses.registered()
 
-        assert mocks_list == responses.mock._matches
+        assert mocks_list == responses.mock.registered()
         assert mocks_list == [first_response, second_response, third_response]
 
     run()
@@ -1792,3 +1946,171 @@ def test_rfc_compliance(url, other_url):
 
     run()
     assert_reset()
+
+
+def test_requests_between_add():
+    @responses.activate
+    def run():
+        responses.add(responses.GET, "https://example.com/", json={"response": "old"})
+        assert requests.get("https://example.com/").content == b'{"response": "old"}'
+        assert requests.get("https://example.com/").content == b'{"response": "old"}'
+        assert requests.get("https://example.com/").content == b'{"response": "old"}'
+
+        responses.add(responses.GET, "https://example.com/", json={"response": "new"})
+
+        assert requests.get("https://example.com/").content == b'{"response": "new"}'
+        assert requests.get("https://example.com/").content == b'{"response": "new"}'
+        assert requests.get("https://example.com/").content == b'{"response": "new"}'
+
+    run()
+    assert_reset()
+
+
+def test_responses_reuse():
+    @responses.activate
+    def run():
+        url = "https://someapi.com/"
+        fail_response = responses.Response(
+            method="GET", url=url, body="fail", status=500
+        )
+        responses.add(responses.GET, url, "success", status=200)
+        responses.add(fail_response)
+        responses.add(fail_response)
+        responses.add(fail_response)
+        responses.add(responses.GET, url, "success", status=200)
+        responses.add(responses.GET, url, "", status=302)
+
+        response = requests.get(url)
+        assert response.content == b"success"
+
+        for _ in range(3):
+            response = requests.get(url)
+            assert response.content == b"fail"
+
+    run()
+    assert_reset()
+
+
+async def test_async_calls():
+    @responses.activate
+    async def run():
+        responses.add(
+            responses.GET,
+            "http://twitter.com/api/1/foobar",
+            json={"error": "not found"},
+            status=404,
+        )
+
+        resp = requests.get("http://twitter.com/api/1/foobar")
+        assert resp.json() == {"error": "not found"}
+        assert responses.calls[0].request.url == "http://twitter.com/api/1/foobar"
+
+    await run()
+    assert_reset()
+
+
+class TestMultipleWrappers:
+    """Test to validate that multiple decorators could be applied.
+
+    Ensures that we can call one function that is wrapped with
+    ``responses.activate`` decorator from within another wrapped function.
+
+    Validates that mock patch is not leaked to other tests.
+    For more detail refer to https://github.com/getsentry/responses/issues/481
+    """
+
+    @responses.activate
+    def test_wrapped(self):
+        responses.add(responses.GET, "http://example.com/1", body="Hello 1")
+        assert b"Hello 1" == requests.get("http://example.com/1").content
+
+    @responses.activate
+    def test_call_another_wrapped_function(self):
+        self.test_wrapped()
+
+    def test_mock_not_leaked(self, httpserver):
+        """
+        Validate that ``responses.activate`` does not leak to unpatched test.
+
+        Parameters
+        ----------
+        httpserver : ContentServer
+            Mock real HTTP server
+
+        """
+        httpserver.serve_content("OK", code=969, headers={"Content-Type": "text/plain"})
+
+        response = requests.get(httpserver.url)
+        assert response.status_code == 969
+
+
+class TestShortcuts:
+    def test_delete(self):
+        @responses.activate
+        def run():
+            responses.delete("http://example.com/1", status=888)
+            resp = requests.delete("http://example.com/1")
+            assert resp.status_code == 888
+
+        run()
+        assert_reset()
+
+    def test_get(self):
+        @responses.activate
+        def run():
+            responses.get("http://example.com/1", status=888)
+            resp = requests.get("http://example.com/1")
+            assert resp.status_code == 888
+
+        run()
+        assert_reset()
+
+    def test_head(self):
+        @responses.activate
+        def run():
+            responses.head("http://example.com/1", status=888)
+            resp = requests.head("http://example.com/1")
+            assert resp.status_code == 888
+
+        run()
+        assert_reset()
+
+    def test_options(self):
+        @responses.activate
+        def run():
+            responses.options("http://example.com/1", status=888)
+            resp = requests.options("http://example.com/1")
+            assert resp.status_code == 888
+
+        run()
+        assert_reset()
+
+    def test_patch(self):
+        @responses.activate
+        def run():
+            responses.patch("http://example.com/1", status=888)
+            resp = requests.patch("http://example.com/1")
+            assert resp.status_code == 888
+
+        run()
+        assert_reset()
+
+    def test_post(self):
+        @responses.activate
+        def run():
+            responses.post("http://example.com/1", status=888)
+            resp = requests.post("http://example.com/1")
+            assert resp.status_code == 888
+
+        run()
+        assert_reset()
+
+    def test_put(self):
+        @responses.activate
+        def run():
+            responses.put("http://example.com/1", status=888)
+            resp = requests.put("http://example.com/1")
+            assert resp.status_code == 888
+
+        run()
+        assert_reset()
